@@ -12,13 +12,13 @@ use std::{
     sync::Arc,
     time::Duration,
 };
-use tracing::{debug, trace};
+use tracing::{trace, warn};
 
 /// How many frames needs to be pending before we start dropping them
 const TERMINAL_FRAMES_DROP: usize = 32;
 
 /// Main trait to interact with a Terminal
-pub trait Terminal: Write {
+pub trait Terminal: Write + Send {
     /// Schedue command for execution
     ///
     /// Command will be submitted on the next call to poll `Terminal::poll`
@@ -105,7 +105,7 @@ pub trait Terminal: Write {
                     let action = handler(self, event, renderer.view())?;
                     // drop frames if we are too far behind
                     if self.frames_pending() > TERMINAL_FRAMES_DROP {
-                        debug!("dropping frames: {}", self.frames_pending());
+                        warn!("dropping frames: {}", self.frames_pending());
                         self.frames_drop();
                         renderer.clear(self)?;
                     }
@@ -242,6 +242,8 @@ pub enum TerminalCommand {
     Char(char),
     /// Set current face (foreground/background colors and text attributes)
     Face(Face),
+    /// Get current face
+    FaceGet,
     /// Control specified DEC mode (DECSET|DECRST)
     DecModeSet { enable: bool, mode: DecMode },
     /// Report specified DEC mode (DECRQM)
@@ -418,12 +420,14 @@ pub enum TerminalEvent {
     Raw(Vec<u8>),
     /// Color
     Color { name: TerminalColor, color: RGBA },
+    /// Report current face
+    FaceGet(Face),
     /// So we can use single decoder for commands and events
     Command(TerminalCommand),
 }
 
 /// Size
-#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Default)]
 pub struct Size {
     pub height: usize,
     pub width: usize,
@@ -445,7 +449,7 @@ impl Size {
 }
 
 /// Terminal size object
-#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Default)]
 pub struct TerminalSize {
     /// Size of the terminal in cells
     pub cells: Size,
