@@ -2,33 +2,33 @@
 use crate::common::IOQueue;
 use crate::encoder::ColorDepth;
 use crate::image::ImageHandlerKind;
-use crate::TerminalCaps;
 use crate::{
     decoder::{Decoder, TTYDecoder},
     encoder::{Encoder, TTYEncoder},
     error::Error,
-    image::{image_handler_detect, DummyImageHandler},
+    image::DummyImageHandler,
     terminal::{
         Size, Terminal, TerminalCommand, TerminalEvent, TerminalSize, TerminalStats, TerminalWaker,
     },
     DecMode, ImageHandler,
 };
+use crate::{KittyImageHandler, SixelImageHandler, TerminalCaps, RGBA};
 use signal_hook::{
     consts::{SIGINT, SIGQUIT, SIGTERM, SIGWINCH},
     iterator::{backend::SignalDelivery, exfiltrator::SignalOnly},
 };
-use std::os::unix::{
-    io::{AsRawFd, RawFd},
-    net::UnixStream,
-};
 use std::{
-    collections::VecDeque,
+    collections::{HashSet, VecDeque},
     fs::File,
     io::{BufWriter, Cursor, Read, Write},
+    os::unix::{
+        io::{AsRawFd, RawFd},
+        net::UnixStream,
+    },
     path::Path,
     time::{Duration, Instant},
 };
-use tracing::info;
+use tracing::{debug, info, warn};
 
 mod nix {
     pub use libc::{winsize, TIOCGWINSZ};
@@ -60,6 +60,9 @@ pub struct UnixTerminal {
     tee: Option<BufWriter<File>>,
     image_handler: Box<dyn ImageHandler + 'static>,
     capabilities: TerminalCaps,
+    // if it is not None we are going to use escape sequence to detect
+    // terminal size, otherwise ioctl is used.
+    size: Option<TerminalSize>,
 }
 
 impl UnixTerminal {
@@ -112,7 +115,7 @@ impl UnixTerminal {
         set_blocking(waker_read.as_raw_fd(), false)?;
 
         let capabilities = TerminalCaps::default();
-        Self {
+        let mut term = Self {
             tty_handle,
             encoder: TTYEncoder::new(capabilities.clone()),
             write_queue: Default::default(),
@@ -126,8 +129,11 @@ impl UnixTerminal {
             tee: None,
             image_handler: Box::new(DummyImageHandler),
             capabilities,
-        }
-        .detect()
+            size: None,
+        };
+
+        capabilities_detect(&mut term)?;
+        Ok(term)
     }
 
     /// Duplicate all output to specified tee file. Used for debugging.
@@ -147,34 +153,29 @@ impl UnixTerminal {
         &mut self.image_handler
     }
 
-    /// Detect termincal capabilieties
-    fn detect(mut self) -> Result<Self, Error> {
-        // color depth
-        let depth = match std::env::var("COLORTERM") {
-            Ok(value) if value == "truecolor" || value == "24bit" => ColorDepth::TrueColor,
-            _ => ColorDepth::EightBit,
-        };
-        let depth = match std::env::var("TERM").as_deref() {
-            Ok("linux") => ColorDepth::Gray,
-            _ => depth,
-        };
-
-        // image handler
-        let image_handler = image_handler_detect(&mut self)?;
-
-        // update capabilieties
-        self.capabilities.glyphs = matches!(
-            image_handler.kind(),
-            ImageHandlerKind::Kitty | ImageHandlerKind::Sixel
-        ) && !self.size()?.pixels.is_empty();
-        self.capabilities.depth = depth;
-
-        // update fields
-        self.encoder = TTYEncoder::new(self.capabilities.clone());
-        self.image_handler = image_handler;
-
-        info!("capabilities: {:?}", self.capabilities);
-        Ok(self)
+    /// Determine terminal size with ioctl
+    ///
+    /// Some terminal emulators do not set pixel size, or if it goes through some
+    /// kind of muxer (like `docker exec`) which might not set pixel size. So if this
+    /// condition is detected we are falling back to escape sequence if it detected to
+    /// work.
+    fn size_ioctl(&self) -> Result<TerminalSize, Error> {
+        unsafe {
+            let mut winsize: nix::winsize = std::mem::zeroed();
+            if libc::ioctl(self.tty_handle.as_raw_fd(), nix::TIOCGWINSZ, &mut winsize) < 0 {
+                return Err(nix::Error::last().into());
+            }
+            Ok(TerminalSize {
+                cells: Size {
+                    height: winsize.ws_row as usize,
+                    width: winsize.ws_col as usize,
+                },
+                pixels: Size {
+                    height: winsize.ws_ypixel as usize,
+                    width: winsize.ws_xpixel as usize,
+                },
+            })
+        }
     }
 
     /// Close all descriptors free all the resources
@@ -234,6 +235,122 @@ impl UnixTerminal {
     }
 }
 
+/// Fallback way to determine terminal size if it is detected to work
+/// and ioctl is not.
+const GET_TERM_SIZE: &[u8] = b"\x1b[18t\x1b[14t";
+
+/// Detect and set terminal capabilities
+fn capabilities_detect(term: &mut UnixTerminal) -> Result<(), Error> {
+    let mut caps = TerminalCaps::default();
+    if let Ok("truecolor") | Ok("24bit") = std::env::var("COLORTERM").as_deref() {
+        caps.depth = ColorDepth::TrueColor;
+    }
+    if let Ok("linux") | Ok("dumb") = std::env::var("TERM").as_deref() {
+        caps.depth = ColorDepth::EightBit;
+    }
+
+    // drain all pending events
+    term.drain().count();
+    // NOTE: using `write!` here instead of execute, to not accidentally use
+    //       existing configuration from passed terminal.
+
+    // 1x1 pixel kitty image (NOTE: it will be consumed by handler if it is already set)
+    write!(term, "\x1b_Ga=q,i=31,s=1,v=1,f=24;AAAA\x1b\\")?;
+
+    // OSC - Get default background color for transparent blending
+    write!(term, "\x1b]11;?\x1b\\")?;
+
+    // Set background color with SGR, and try to get it back to
+    // detect truecolor support https://github.com/termstandard/colors
+    let face_expected = "bg=#010203".parse()?;
+    write!(term, "\x1b[00;48;2;1;2;3m")?; // change background
+    write!(term, "\x1bP$qm\x1b\\")?; // DECRQSS with `m` descriptor
+    write!(term, "\x1b[00m")?; // reset current face
+
+    // Detect terminal size
+    // Some terminals return incomplete size info with ioctl
+    term.write_all(GET_TERM_SIZE)?;
+
+    // DA1 - sync and sixel info
+    // Device Attribute comand is used as "sync" event, it is supported
+    // by most terminals, at least in its basic form, so we expect to
+    // receive a response to it. Which means it should go LAST
+    write!(term, "\x1b[c")?;
+
+    let mut image_handlers = HashSet::new();
+    let mut bg: Option<RGBA> = None;
+    let mut size_escape = TerminalSize::default();
+    loop {
+        match term.poll(Some(Duration::from_secs(1)))? {
+            Some(TerminalEvent::KittyImage { .. }) => {
+                debug!("[detected] kitty image protocol");
+                image_handlers.insert(ImageHandlerKind::Kitty);
+            }
+            Some(TerminalEvent::Color { color, .. }) => {
+                debug!("[detected] background color: {:?}", color);
+                bg.replace(color);
+            }
+            Some(TerminalEvent::FaceGet(face)) => {
+                if face == face_expected {
+                    debug!("[detected] true color support");
+                    caps.depth = ColorDepth::TrueColor;
+                }
+            }
+            Some(TerminalEvent::DeviceAttrs(attrs)) => {
+                // 4 - attribute indicates sixel support
+                if attrs.contains(&4) {
+                    debug!("[detected] sixel image protocol");
+                    image_handlers.insert(ImageHandlerKind::Sixel);
+                }
+                break; // this is last "sync" event
+            }
+            Some(TerminalEvent::Size(size)) => {
+                size_escape = size;
+            }
+            Some(event) => {
+                warn!("unexpected event during detection: {:?}", event);
+                continue;
+            }
+            None => break,
+        }
+    }
+
+    // drain terminal
+    term.drain().count();
+
+    // term size interface
+    let size_ioctl = term.size_ioctl()?;
+    if size_ioctl.pixels.is_empty() && !size_escape.pixels.is_empty() {
+        warn!("[detect] fallback to escape sequence for term size detection");
+        term.size = Some(size_escape);
+    }
+
+    // image handler
+    let image_handler: Box<dyn ImageHandler> = {
+        if image_handlers.contains(&ImageHandlerKind::Kitty) {
+            Box::new(KittyImageHandler::new())
+        } else if image_handlers.contains(&ImageHandlerKind::Sixel) {
+            Box::new(SixelImageHandler::new(bg))
+        } else {
+            Box::new(DummyImageHandler)
+        }
+    };
+
+    // glyph support
+    caps.glyphs = matches!(
+        image_handler.kind(),
+        ImageHandlerKind::Kitty | ImageHandlerKind::Sixel
+    ) && !term.size()?.pixels.is_empty();
+
+    // update terminal
+    info!("capabilities: {:?}", caps);
+    term.encoder = TTYEncoder::new(caps.clone());
+    term.image_handler = image_handler;
+    term.capabilities = caps;
+
+    Ok(())
+}
+
 impl std::ops::Drop for UnixTerminal {
     fn drop(&mut self) {
         self.dispose().unwrap_or(())
@@ -315,8 +432,12 @@ impl Terminal for UnixTerminal {
                 for signal in self.signal_delivery.pending() {
                     match signal {
                         SIGWINCH => {
-                            self.events_queue
-                                .push_back(TerminalEvent::Resize(self.size()?));
+                            if self.size.is_none() {
+                                self.events_queue
+                                    .push_back(TerminalEvent::Resize(self.size()?));
+                            } else {
+                                self.write_all(GET_TERM_SIZE)?;
+                            }
                         }
                         SIGTERM | SIGINT | SIGQUIT => {
                             return Err(Error::Quit);
@@ -343,6 +464,13 @@ impl Terminal for UnixTerminal {
                 // parse events
                 let mut read_queue = Cursor::new(&buf[..recv]);
                 while let Some(event) = self.decoder.decode(&mut read_queue)? {
+                    if let TerminalEvent::Size(size) = event {
+                        // we are using escape sequence to determine terminal resize
+                        if let Some(term_size) = self.size.as_mut() {
+                            *term_size = size;
+                            self.events_queue.push_back(TerminalEvent::Resize(size));
+                        }
+                    }
                     if !self.image_handler.handle(&event)? {
                         self.events_queue.push_back(event)
                     }
@@ -374,21 +502,9 @@ impl Terminal for UnixTerminal {
     }
 
     fn size(&self) -> Result<TerminalSize, Error> {
-        unsafe {
-            let mut winsize: nix::winsize = std::mem::zeroed();
-            if libc::ioctl(self.tty_handle.as_raw_fd(), nix::TIOCGWINSZ, &mut winsize) < 0 {
-                return Err(nix::Error::last().into());
-            }
-            Ok(TerminalSize {
-                cells: Size {
-                    height: winsize.ws_row as usize,
-                    width: winsize.ws_col as usize,
-                },
-                pixels: Size {
-                    height: winsize.ws_ypixel as usize,
-                    width: winsize.ws_xpixel as usize,
-                },
-            })
+        match self.size {
+            Some(size) => Ok(size),
+            None => self.size_ioctl(),
         }
     }
 
